#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2007 Donald N. Allingham
# Copyright (C) 2008      Lukasz Rymarczyk
# Copyright (C) 2008      Raphael Ackermann
# Copyright (C) 2008-2011 Brian G. Matherly
# Copyright (C) 2010      Jakim Friant
# Copyright (C) 2011-2013 Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#
# cli.plug.__init__
#

"""Enable report generation from the command line interface (CLI)"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import traceback
import os
import sys

import logging

LOG = logging.getLogger(".")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.plug import BasePluginManager
from gramps.gen.plug.docgen import (
    StyleSheet,
    StyleSheetList,
    PaperStyle,
    PAPER_PORTRAIT,
    PAPER_LANDSCAPE,
    graphdoc,
    treedoc,
)
from gramps.gen.plug.menu import (
    FamilyOption,
    PersonOption,
    NoteOption,
    MediaOption,
    PersonListOption,
    NumberOption,
    BooleanOption,
    DestinationOption,
    Option,
    TextOption,
    EnumeratedListOption,
    StringOption,
)
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.errors import ReportError, FilterError
from gramps.gen.plug.report import (
    CATEGORY_TEXT,
    CATEGORY_DRAW,
    CATEGORY_BOOK,
    CATEGORY_GRAPHVIZ,
    CATEGORY_TREE,
    CATEGORY_CODE,
    ReportOptions,
    append_styles,
)
from gramps.gen.plug.report._paper import paper_sizes
from gramps.gen.const import USER_HOME, DOCGEN_OPTIONS
from gramps.gen.dbstate import DbState
from ..grampscli import CLIManager
from ..user import User
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# ------------------------------------------------------------------------
#
# Private Functions
#
# ------------------------------------------------------------------------
def _convert_str_to_match_type(str_val, type_val):
    """
    Returns a value representing str_val that is the same type as type_val.
    """
    str_val = str_val.strip()
    ret_type = type(type_val)

    if isinstance(type_val, str):
        if (str_val.startswith("'") and str_val.endswith("'")) or (
            str_val.startswith('"') and str_val.endswith('"')
        ):
            # Remove enclosing quotes
            return str(str_val[1:-1])
        else:
            return str(str_val)

    elif ret_type == int:
        try:
            return int(str_val)
        except ValueError:
            print("'%s' is not an integer number" % str_val)
            return 0

    elif ret_type == float:
        try:
            return float(str_val)
        except ValueError:
            print("'%s' is not a decimal number" % str_val)
            return 0.0

    elif ret_type == bool:
        if str_val == str(True):
            return True
        elif str_val == str(False):
            return False
        else:
            print("'%s' is not a boolean-- try 'True' or 'False'" % str_val)
            return False

    elif ret_type == list:
        ret_val = []
        if not (str_val.startswith("[") and str_val.endswith("]")):
            print("'%s' is not a list-- try: [%s]" % (str_val, str_val))
            return ret_val

        entry = ""
        quote_type = None

        # Search through characters between the brackets
        for char in str_val[1:-1]:
            if (char == "'" or char == '"') and quote_type is None:
                # This character starts a string
                quote_type = char
            elif char == quote_type:
                # This character ends a string
                quote_type = None
            elif quote_type is None and char == ",":
                # This character ends an entry
                ret_val.append(entry.strip())
                entry = ""
                quote_type = None
            else:
                entry += char

        if entry != "":
            # Add the last entry
            ret_val.append(entry.strip())

        return ret_val


def _validate_options(options, dbase):
    """
    Validate all options by making sure that their values are consistent with
    the database.

    menu: The Menu class
    dbase: the database the options will be applied to
    """
    if not hasattr(options, "menu"):
        return
    menu = options.menu

    for name in menu.get_all_option_names():
        option = menu.get_option_by_name(name)

        if isinstance(option, PersonOption):
            pid = option.get_value()
            person = dbase.get_person_from_gramps_id(pid)
            if not person:
                person = dbase.get_default_person()
                if not person:
                    try:
                        phandle = next(dbase.iter_person_handles())
                    except StopIteration:
                        phandle = None
                    person = dbase.get_person_from_handle(phandle)
                    if not person:
                        print(_("ERROR: Please specify a person"), file=sys.stderr)
            if person:
                option.set_value(person.get_gramps_id())

        elif isinstance(option, FamilyOption):
            fid = option.get_value()
            family = dbase.get_family_from_gramps_id(fid)
            if not family:
                person = dbase.get_default_person()
                family_list = []
                family_handle = None
                if person:
                    family_list = person.get_family_handle_list()
                if family_list:
                    family_handle = family_list[0]
                else:
                    try:
                        family_handle = next(dbase.iter_family_handles())
                    except StopIteration:
                        family_handle = None
                if family_handle:
                    family = dbase.get_family_from_handle(family_handle)
                    option.set_value(family.get_gramps_id())
                else:
                    print(_("ERROR: Please specify a family"), file=sys.stderr)


# ------------------------------------------------------------------------
#
# Command-line report
#
# ------------------------------------------------------------------------
class CommandLineReport:
    """
    Provide a way to generate report from the command line.
    """

    def __init__(
        self, database, name, category, option_class, options_str_dict, noopt=False
    ):
        pmgr = BasePluginManager.get_instance()
        self.__textdoc_plugins = []
        self.__drawdoc_plugins = []
        self.__bookdoc_plugins = []
        for plugin in pmgr.get_docgen_plugins():
            if plugin.get_text_support() and plugin.get_extension():
                self.__textdoc_plugins.append(plugin)
            if plugin.get_draw_support() and plugin.get_extension():
                self.__drawdoc_plugins.append(plugin)
            if (
                plugin.get_extension()
                and plugin.get_text_support()
                and plugin.get_draw_support()
            ):
                self.__bookdoc_plugins.append(plugin)

        self.database = database
        self.category = category

        self.options_dict = None  # keep pylint happy
        self.options_help = None
        self.paper = None
        self.orien = None
        self.marginl = None
        self.marginr = None
        self.margint = None
        self.marginb = None
        self.doc_options = None
        self.doc_option_class = None
        self.selected_style = None
        self.style_list = None
        self.css_filename = None
        self.format = None
        self.raw_name = name

        self.option_class = option_class(name, database)
        if category == CATEGORY_GRAPHVIZ:
            # Need to include Graphviz options
            self.__gvoptions = graphdoc.GVOptions()
            menu = self.option_class.menu
            self.__gvoptions.add_menu_options(menu)
            for name in menu.get_all_option_names():
                if name not in self.option_class.options_dict:
                    self.option_class.options_dict[name] = menu.get_option_by_name(
                        name
                    ).get_value()
        if category == CATEGORY_TREE:
            # Need to include Genealogy Tree options
            self.__toptions = treedoc.TreeOptions()
            menu = self.option_class.menu
            self.__toptions.add_menu_options(menu)
            for name in menu.get_all_option_names():
                if name not in self.option_class.options_dict:
                    self.option_class.options_dict[name] = menu.get_option_by_name(
                        name
                    ).get_value()
        self.option_class.load_previous_values()
        _validate_options(self.option_class, database)
        self.show = options_str_dict.pop("show", None)

        self.options_str_dict = options_str_dict
        self.init_standard_options(noopt)
        self.init_report_options()
        self.parse_options()
        self.init_report_options_help()
        self.show_options()

    def init_standard_options(self, noopt):
        """
        Initialize the options that are hard-coded into the report system.
        """
        self.options_dict = {
            "of": self.option_class.handler.module_name,
            "off": self.option_class.handler.get_format_name(),
            "style": self.option_class.handler.get_default_stylesheet_name(),
            "papers": self.option_class.handler.get_paper_name(),
            "papero": self.option_class.handler.get_orientation(),
            "paperml": self.option_class.handler.get_margins()[0],
            "papermr": self.option_class.handler.get_margins()[1],
            "papermt": self.option_class.handler.get_margins()[2],
            "papermb": self.option_class.handler.get_margins()[3],
            "css": self.option_class.handler.get_css_filename(),
        }

        self.options_help = {
            "of": [_("=filename"), _("Output file name. MANDATORY"), ""],
            "off": [_("=format"), _("Output file format."), []],
            "style": [_("=name"), _("Style name."), ""],
            "papers": [_("=name"), _("Paper size name."), ""],
            "papero": [_("=number"), _("Paper orientation number."), ""],
            "paperml": [_("=number"), _("Left paper margin"), _("Size in cm")],
            "papermr": [_("=number"), _("Right paper margin"), _("Size in cm")],
            "papermt": [_("=number"), _("Top paper margin"), _("Size in cm")],
            "papermb": [_("=number"), _("Bottom paper margin"), _("Size in cm")],
            "css": [_("=css filename"), _("CSS filename to use, html format only"), ""],
        }

        if noopt:
            return

        self.options_help["of"][2] = os.path.join(USER_HOME, "whatever_name")

        if self.category == CATEGORY_TEXT:
            for plugin in self.__textdoc_plugins:
                self.options_help["off"][2].append(
                    plugin.get_extension() + "\t" + plugin.get_description()
                )
        elif self.category == CATEGORY_DRAW:
            for plugin in self.__drawdoc_plugins:
                self.options_help["off"][2].append(
                    plugin.get_extension() + "\t" + plugin.get_description()
                )
        elif self.category == CATEGORY_BOOK:
            for plugin in self.__bookdoc_plugins:
                self.options_help["off"][2].append(
                    plugin.get_extension() + "\t" + plugin.get_description()
                )
        elif self.category == CATEGORY_GRAPHVIZ:
            for graph_format in graphdoc.FORMATS:
                self.options_help["off"][2].append(
                    graph_format["type"] + "\t" + graph_format["descr"]
                )
        elif self.category == CATEGORY_TREE:
            for tree_format in treedoc.FORMATS:
                self.options_help["off"][2].append(
                    tree_format["type"] + "\t" + tree_format["descr"]
                )
        else:
            self.options_help["off"][2] = "NA"

        self.options_help["papers"][2] = [
            paper.get_name()
            for paper in paper_sizes
            if paper.get_name() != "Custom Size"
        ]

        self.options_help["papero"][2] = [
            "%d\tPortrait" % PAPER_PORTRAIT,
            "%d\tLandscape" % PAPER_LANDSCAPE,
        ]

        self.options_help["css"][2] = os.path.join(USER_HOME, "whatever_name.css")

        if self.category in (CATEGORY_TEXT, CATEGORY_DRAW):
            default_style = StyleSheet()
            self.option_class.make_default_style(default_style)

            # Read all style sheets available for this item
            style_file = self.option_class.handler.get_stylesheet_savefile()
            self.style_list = StyleSheetList(style_file, default_style)

            self.options_help["style"][2] = self.style_list.get_style_names()

    def init_report_options(self):
        """
        Initialize the options that are defined by each report.
        """

        if self.category == CATEGORY_BOOK:  # a Book Report has no "menu"
            for key in self.option_class.options_dict:
                self.options_dict[key] = self.option_class.options_dict[key]
                self.options_help[key] = self.option_class.options_help[key][:3]
            # a Book Report can't have HTML output so "css" is meaningless
            self.options_dict.pop("css")

        if not hasattr(self.option_class, "menu"):
            return
        menu = self.option_class.menu
        for name in menu.get_all_option_names():
            option = menu.get_option_by_name(name)
            self.options_dict[name] = option.get_value()

    def init_report_options_help(self):
        """
        Initialize help for the options that are defined by each report.
        (And also any docgen options, if defined by the docgen.)
        """
        if not hasattr(self.option_class, "menu"):
            return
        menu = self.option_class.menu
        for name in menu.get_all_option_names():
            option = menu.get_option_by_name(name)
            self.options_help[name] = ["", option.get_help()]

            if isinstance(option, PersonOption):
                id_list = []
                for person_handle in self.database.get_person_handles(True):
                    person = self.database.get_person_from_handle(person_handle)
                    id_list.append(
                        "%s\t%s"
                        % (person.get_gramps_id(), name_displayer.display(person))
                    )
                self.options_help[name].append(id_list)
            elif isinstance(option, FamilyOption):
                id_list = []
                for family in self.database.iter_families():
                    mname = ""
                    fname = ""
                    mhandle = family.get_mother_handle()
                    if mhandle:
                        mother = self.database.get_person_from_handle(mhandle)
                        if mother:
                            mname = name_displayer.display(mother)
                    fhandle = family.get_father_handle()
                    if fhandle:
                        father = self.database.get_person_from_handle(fhandle)
                        if father:
                            fname = name_displayer.display(father)
                    # Translators: needed for French, Hebrew and Arabic
                    text = _("%(id)s:\t%(father)s, %(mother)s") % {
                        "id": family.get_gramps_id(),
                        "father": fname,
                        "mother": mname,
                    }
                    id_list.append(text)
                self.options_help[name].append(id_list)
            elif isinstance(option, NoteOption):
                id_list = []
                for nhandle in self.database.get_note_handles():
                    note = self.database.get_note_from_handle(nhandle)
                    id_list.append(note.get_gramps_id())
                self.options_help[name].append(id_list)
            elif isinstance(option, MediaOption):
                id_list = []
                for mhandle in self.database.get_media_handles():
                    mobject = self.database.get_media_from_handle(mhandle)
                    id_list.append(mobject.get_gramps_id())
                self.options_help[name].append(id_list)
            elif isinstance(option, PersonListOption):
                self.options_help[name].append("")
            elif isinstance(option, NumberOption):
                self.options_help[name].append("A number")
            elif isinstance(option, BooleanOption):
                self.options_help[name].append(["False", "True"])
            elif isinstance(option, DestinationOption):
                self.options_help[name].append("A file system path")
            elif isinstance(option, StringOption):
                self.options_help[name].append("Any text")
            elif isinstance(option, TextOption):
                self.options_help[name].append(
                    "A list of text values. Each entry in the list "
                    "represents one line of text."
                )
            elif isinstance(option, EnumeratedListOption):
                ilist = []
                for value, description in option.get_items():
                    tabs = "\t"
                    try:
                        tabs = "\t\t" if len(value) < 10 else "\t"
                    except TypeError:  # Value is a number, use just one tab.
                        pass
                    val = "%s%s%s" % (value, tabs, description)
                    ilist.append(val)
                self.options_help[name].append(ilist)
            elif isinstance(option, Option):
                self.options_help[name].append(option.get_help())
            else:
                print(_("Unknown option: %s") % option, file=sys.stderr)
                print(
                    _("   Valid options are:")
                    + _(", ").join(list(self.options_dict.keys())),  # Arabic OK
                    file=sys.stderr,
                )
                print(
                    _(
                        "   Use '%(donottranslate)s' to see description "
                        "and acceptable values"
                    )
                    % {"donottranslate": "show=option"},
                    file=sys.stderr,
                )

    def parse_options(self):
        """
        Load the options that the user has entered.
        """
        if not hasattr(self.option_class, "menu"):
            menu = None
        else:
            menu = self.option_class.menu
            menu_opt_names = menu.get_all_option_names()

        _format_str = self.options_str_dict.pop("off", None)
        if _format_str:
            self.options_dict["off"] = _format_str

        self.css_filename = None
        _chosen_format = None

        self.doc_option_class = None
        if self.category in [CATEGORY_TEXT, CATEGORY_DRAW, CATEGORY_BOOK]:
            if self.category == CATEGORY_TEXT:
                plugins = self.__textdoc_plugins
                self.css_filename = self.options_dict["css"]
            elif self.category == CATEGORY_DRAW:
                plugins = self.__drawdoc_plugins
            elif self.category == CATEGORY_BOOK:
                plugins = self.__bookdoc_plugins
            for plugin in plugins:
                if plugin.get_extension() == self.options_dict["off"]:
                    self.format = plugin.get_basedoc()
                    self.doc_option_class = plugin.get_doc_option_class()
            if self.format is None:
                # Pick the first one as the default.
                plugin = plugins[0]
                self.format = plugin.get_basedoc()
                self.doc_option_class = plugin.get_doc_option_class()
                _chosen_format = plugin.get_extension()
        elif self.category == CATEGORY_GRAPHVIZ:
            for graph_format in graphdoc.FORMATS:
                if graph_format["type"] == self.options_dict["off"]:
                    if not self.format:  # choose the first one, not the last
                        self.format = graph_format["class"]
            if self.format is None:
                # Pick the first one as the default.
                self.format = graphdoc.FORMATS[0]["class"]
                _chosen_format = graphdoc.FORMATS[0]["type"]
        elif self.category == CATEGORY_TREE:
            for tree_format in treedoc.FORMATS:
                if tree_format["type"] == self.options_dict["off"]:
                    if not self.format:  # choose the first one, not the last
                        self.format = tree_format["class"]
            if self.format is None:
                # Pick the first one as the default.
                self.format = treedoc.FORMATS[0]["class"]
                _chosen_format = treedoc.FORMATS[0]["type"]
        else:
            self.format = None
        if _chosen_format and _format_str:
            print(
                _(
                    "Ignoring '%(notranslate1)s=%(notranslate2)s' "
                    "and using '%(notranslate1)s=%(notranslate3)s'."
                )
                % {
                    "notranslate1": "off",
                    "notranslate2": self.options_dict["off"],
                    "notranslate3": _chosen_format,
                },
                file=sys.stderr,
            )
            print(
                _("Use '%(notranslate)s' to see valid values.")
                % {"notranslate": "show=off"},
                file=sys.stderr,
            )

        self.do_doc_options()

        for opt in self.options_str_dict:
            if opt in self.options_dict:
                self.options_dict[opt] = _convert_str_to_match_type(
                    self.options_str_dict[opt], self.options_dict[opt]
                )

                self.option_class.handler.options_dict[opt] = self.options_dict[opt]

                if menu and opt in menu_opt_names:
                    option = menu.get_option_by_name(opt)
                    option.set_value(self.options_dict[opt])

            else:
                print(_("Ignoring unknown option: %s") % opt, file=sys.stderr)
                print(
                    _("   Valid options are:"),
                    _(", ").join(list(self.options_dict.keys())),  # Arabic OK
                    file=sys.stderr,
                )
                print(
                    _(
                        "   Use '%(donottranslate)s' to see description "
                        "and acceptable values"
                    )
                    % {"donottranslate": "show=option"},
                    file=sys.stderr,
                )

        self.option_class.handler.output = self.options_dict["of"]

        self.paper = paper_sizes[0]  # make sure one exists
        for paper in paper_sizes:
            if paper.get_name() == self.options_dict["papers"]:
                self.paper = paper
        self.option_class.handler.set_paper(self.paper)

        self.orien = self.options_dict["papero"]

        self.marginl = self.options_dict["paperml"]
        self.marginr = self.options_dict["papermr"]
        self.margint = self.options_dict["papermt"]
        self.marginb = self.options_dict["papermb"]

        if self.category in (CATEGORY_TEXT, CATEGORY_DRAW):
            default_style = StyleSheet()
            self.option_class.make_default_style(default_style)

            # Read all style sheets available for this item
            style_file = self.option_class.handler.get_stylesheet_savefile()
            self.style_list = StyleSheetList(style_file, default_style)

            # Get the selected stylesheet
            style_name = self.option_class.handler.get_default_stylesheet_name()
            self.selected_style = self.style_list.get_style_sheet(style_name)

    def do_doc_options(self):
        """
        Process docgen options, if any (options for the backend, e.g. AsciiDoc)
        """
        self.doc_options = None
        if not self.doc_option_class:
            return  # this docgen type has no options
        try:
            if issubclass(self.doc_option_class, object):
                self.doc_options = self.doc_option_class(self.raw_name, self.database)
                doc_options_dict = self.doc_options.options_dict
        except TypeError:
            self.doc_options = self.doc_option_class
        self.doc_options.load_previous_values()
        docgen_menu = self.doc_options.menu
        report_menu = self.option_class.menu  # "help" checks the option type
        for oname in docgen_menu.get_option_names(DOCGEN_OPTIONS):
            docgen_opt = docgen_menu.get_option(DOCGEN_OPTIONS, oname)
            if oname in self.options_str_dict and oname in doc_options_dict:
                doc_options_dict[oname] = _convert_str_to_match_type(
                    self.options_str_dict[oname], doc_options_dict[oname]
                )
                self.options_str_dict.pop(oname)
            if oname in doc_options_dict:
                docgen_opt.set_value(doc_options_dict[oname])
            report_menu.add_option(DOCGEN_OPTIONS, oname, docgen_opt)
        for oname in doc_options_dict:  # enable "help"
            self.options_dict[oname] = doc_options_dict[oname]
            self.options_help[oname] = self.doc_options.options_help[oname][:3]

    def show_options(self):
        """
        Print available options on the CLI.
        """
        if not self.show:
            return
        elif self.show == "all":
            print(_("   Available options:"))
            for key in sorted(self.options_dict.keys()):
                if key in self.options_help:
                    opt = self.options_help[key]
                    # Make the output nicer to read, assume a tab has 8 spaces
                    tabs = "\t\t" if len(key) < 10 else "\t"
                    optmsg = "      %s%s%s (%s)" % (key, tabs, opt[1], opt[0])
                else:
                    optmsg = "      %s%s%s" % (key, tabs, _("(no help available)"))
                print(optmsg)
            print(
                _(
                    "   Use '%(donottranslate)s' to see description "
                    "and acceptable values"
                )
                % {"donottranslate": "show=option"}
            )
        elif self.show in self.options_help:
            opt = self.options_help[self.show]
            tabs = "\t\t" if len(self.show) < 10 else "\t"
            print(_("   Available values are:"))
            print("      %s%s%s (%s)" % (self.show, tabs, opt[1], opt[0]))
            vals = opt[2]
            if isinstance(vals, (list, tuple)):
                for val in vals:
                    print("      %s" % val)
            else:
                print("      %s" % opt[2])

        else:
            # there was a show option given, but the option is invalid
            print(
                _(
                    "option '%(optionname)s' not valid. "
                    "Use '%(donottranslate)s' to see all valid options."
                )
                % {"optionname": self.show, "donottranslate": "show=all"},
                file=sys.stderr,
            )


# ------------------------------------------------------------------------
#
# Command-line report generic task
#
# ------------------------------------------------------------------------
def cl_report(database, name, category, report_class, options_class, options_str_dict):
    """
    function to actually run the selected report
    """

    err_msg = _("Failed to write report. ")
    clr = CommandLineReport(database, name, category, options_class, options_str_dict)

    # Exit here if show option was given
    if clr.show:
        return

    # write report
    try:
        if category in [CATEGORY_TEXT, CATEGORY_DRAW, CATEGORY_BOOK]:
            if clr.doc_options:
                clr.option_class.handler.doc = clr.format(
                    clr.selected_style,
                    PaperStyle(
                        clr.paper,
                        clr.orien,
                        clr.marginl,
                        clr.marginr,
                        clr.margint,
                        clr.marginb,
                    ),
                    clr.doc_options,
                )
            else:
                clr.option_class.handler.doc = clr.format(
                    clr.selected_style,
                    PaperStyle(
                        clr.paper,
                        clr.orien,
                        clr.marginl,
                        clr.marginr,
                        clr.margint,
                        clr.marginb,
                    ),
                )
        elif category in [CATEGORY_GRAPHVIZ, CATEGORY_TREE]:
            clr.option_class.handler.doc = clr.format(
                clr.option_class,
                PaperStyle(
                    clr.paper,
                    clr.orien,
                    clr.marginl,
                    clr.marginr,
                    clr.margint,
                    clr.marginb,
                ),
            )
        if clr.css_filename is not None and hasattr(
            clr.option_class.handler.doc, "set_css_filename"
        ):
            clr.option_class.handler.doc.set_css_filename(clr.css_filename)
        my_report = report_class(database, clr.option_class, User())
        my_report.doc.init()
        my_report.begin_report()
        my_report.write_report()
        my_report.end_report()
        return clr
    except ReportError as msg:
        (msg1, msg2) = msg.messages()
        print(err_msg, file=sys.stderr)
        print(msg1, file=sys.stderr)
        if msg2:
            print(msg2, file=sys.stderr)
    except:
        if len(LOG.handlers) > 0:
            LOG.error(err_msg, exc_info=True)
        else:
            print(err_msg, file=sys.stderr)
            ## Something seems to eat the exception above.
            ## Hack to re-get the exception:
            try:
                raise
            except:
                traceback.print_exc()


def run_report(db, name, **options_str_dict):
    """
    Given a database, run a given report.

    db is a Db database

    name is the name of a report

    options_str_dict is the same kind of options
    given at the command line. For example:

    >>> run_report(db, "ancestor_report", off="txt",
                   of="ancestor-007.txt", pid="I37")

    returns CommandLineReport (clr) if successfully runs the report,
    None otherwise.

    You can see:
       options and values used in  clr.option_class.options_dict
       filename in clr.option_class.get_output()
    """
    dbstate = DbState()
    climanager = CLIManager(dbstate, False, User())  # don't load db
    climanager.do_reg_plugins(dbstate, None)
    pmgr = BasePluginManager.get_instance()
    cl_list = pmgr.get_reg_reports()
    clr = None
    for pdata in cl_list:
        if name == pdata.id:
            mod = pmgr.load_plugin(pdata)
            if not mod:
                # import of plugin failed
                return clr
            category = pdata.category
            report_class = getattr(mod, pdata.reportclass)
            options_class = getattr(mod, pdata.optionclass)
            if category in (CATEGORY_BOOK, CATEGORY_CODE):
                options_class(db, name, category, options_str_dict)
            else:
                clr = cl_report(
                    db, name, category, report_class, options_class, options_str_dict
                )
                return clr
    return clr


# ------------------------------------------------------------------------
#
# Function to write books from command line
#
# ------------------------------------------------------------------------
def cl_book(database, name, book, options_str_dict):
    """
    function to actually run the selected book,
    which in turn runs whatever reports the book has in it
    """

    clr = CommandLineReport(
        database, name, CATEGORY_BOOK, ReportOptions, options_str_dict
    )

    # Exit here if show option was given
    if clr.show:
        return

    # write report
    doc = clr.format(
        None,
        PaperStyle(
            clr.paper, clr.orien, clr.marginl, clr.marginr, clr.margint, clr.marginb
        ),
    )
    user = User()
    rptlist = []
    selected_style = StyleSheet()
    for item in book.get_item_list():
        # The option values were loaded magically by the book parser.
        # But they still need to be applied to the menu options.
        opt_dict = item.option_class.options_dict
        menu = item.option_class.menu
        for optname in opt_dict:
            menu_option = menu.get_option_by_name(optname)
            if menu_option:
                menu_option.set_value(opt_dict[optname])

        item.option_class.set_document(doc)
        report_class = item.get_write_item()
        obj = (
            write_book_item(database, report_class, item.option_class, user),
            item.get_translated_name(),
        )
        if obj:
            append_styles(selected_style, item)
            rptlist.append(obj)

    doc.set_style_sheet(selected_style)
    doc.open(clr.option_class.get_output())
    doc.init()
    newpage = 0
    err_msg = _("Failed to make '%s' report.")
    try:
        for rpt, name in rptlist:
            if newpage:
                doc.page_break()
            newpage = 1
            rpt.begin_report()
            rpt.write_report()
        doc.close()
    except ReportError as msg:
        (msg1, msg2) = msg.messages()
        print(err_msg % name, file=sys.stderr)  # which report has the error?
        print(msg1, file=sys.stderr)
        if msg2:
            print(msg2, file=sys.stderr)


# ------------------------------------------------------------------------
#
# Generic task function for book
#
# ------------------------------------------------------------------------
def write_book_item(database, report_class, options, user):
    """Write the report using options set.
    All user dialog has already been handled and the output file opened."""
    try:
        return report_class(database, options, user)
    except ReportError as msg:
        (msg1, msg2) = msg.messages()
        print("ReportError", msg1, msg2, file=sys.stderr)
    except FilterError as msg:
        (msg1, msg2) = msg.messages()
        print("FilterError", msg1, msg2, file=sys.stderr)
    except:
        LOG.error("Failed to write book item.", exc_info=True)
    return None



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008  Zsolt Foldvari
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


"Text editor subclassed from Gtk.TextView handling :class:`.StyledText`."

__all__ = ["StyledTextEditor"]

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

import logging

_LOG = logging.getLogger(".widgets.styledtexteditor")

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango
from gi.repository.Gio import SimpleActionGroup
from gi.repository.GLib import Variant

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.lib import StyledTextTagType
from .styledtextbuffer import (
    ALLOWED_STYLES,
    MATCH_START,
    MATCH_END,
    MATCH_FLAVOR,
    MATCH_STRING,
    LinkTag,
)
from .undoablestyledbuffer import UndoableStyledBuffer
from ..spell import Spell
from ..display import display_url
from ..utils import SystemFonts, match_primary_mask, get_link_color
from gramps.gen.config import config
from gramps.gen.constfunc import has_display, mac
from ..uimanager import ActionGroup

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
if has_display():
    display = Gdk.Display.get_default()
    HAND_CURSOR = Gdk.Cursor.new_for_display(display, Gdk.CursorType.HAND2)
    REGULAR_CURSOR = Gdk.Cursor.new_for_display(display, Gdk.CursorType.XTERM)

FORMAT_TOOLBAR = (
    """<?xml version="1.0" encoding="UTF-8"?>
<interface>
  <object class="GtkToolbar" id="ToolBar">
    <property name="hexpand">True</property>
    <property name="toolbar-style">GTK_TOOLBAR_ICONS</property>
    <style>
      <class name="primary-toolbar"/>
    </style>
    <child>
      <object class="GtkToggleToolButton">
        <property name="icon-name">format-text-italic</property>
        <property name="action-name">ste.ITALIC</property>
        <property name="tooltip_text" translatable="yes">Italic</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child>
      <object class="GtkToggleToolButton">
        <property name="icon-name">format-text-bold</property>
        <property name="action-name">ste.BOLD</property>
        <property name="tooltip_text" translatable="yes">Bold</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child>
      <object class="GtkToggleToolButton">
        <property name="icon-name">format-text-underline</property>
        <property name="action-name">ste.UNDERLINE</property>
        <property name="tooltip_text" translatable="yes">Underline</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child>
      <object class="GtkToggleToolButton">
        <property name="icon-name">format-text-strikethrough</property>
        <property name="action-name">ste.STRIKETHROUGH</property>
        <property name="tooltip_text" translatable="yes">Strikethrough</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child>
      <object class="GtkToggleToolButton" id="superscript">
        <property name="icon-name">format-text-superscript</property>
        <property name="action-name">ste.SUPERSCRIPT</property>
        <property name="tooltip_text" translatable="yes">Superscript</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child>
      <object class="GtkToggleToolButton" id="subscript">
        <property name="icon-name">format-text-subscript</property>
        <property name="action-name">ste.SUBSCRIPT</property>
        <property name="tooltip_text" translatable="yes">Subscript</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child>
      <object class="GtkToolItem">
        <property name="tooltip_text" translatable="yes">Font family</property>
        <child>
          <object class="ShortlistComboEntry" id="Fontface"></object>
        </child>
      </object>
    </child>
    <child>
      <object class="GtkToolItem">
        <property name="tooltip_text" translatable="yes">Font size</property>
        <child>
          <object class="ShortlistComboEntry" id="Fontsize"></object>
        </child>
      </object>
    </child>
    <child>
      <object class="GtkToolButton">
        <property name="icon-name">edit-undo</property>
        <property name="action-name">ste.STUndo</property>
        <property name="tooltip_text" translatable="yes">Undo</property>
        <property name="label" translatable="yes">Undo</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child>
      <object class="GtkToolButton">
        <property name="icon-name">edit-redo</property>
        <property name="action-name">ste.STRedo</property>
        <property name="tooltip_text" translatable="yes">Redo</property>
        <property name="label" translatable="yes">Redo</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child>
      <object class="GtkToolButton">
        <property name="icon-name">gramps-font-color</property>
        <property name="action-name">ste.FONTCOLOR</property>
        <property name="tooltip_text" translatable="yes">Font Color</property>
        <property name="label" translatable="yes">Font Color</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child>
      <object class="GtkToolButton">
        <property name="icon-name">gramps-font-bgcolor</property>
        <property name="action-name">ste.HIGHLIGHT</property>
        <property name="tooltip_text" translatable="yes">"""
    """Background Color</property>
        <property name="label" translatable="yes">Background Color</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child>
      <object class="GtkToolButton">
        <property name="icon-name">go-jump</property>
        <property name="action-name">ste.LINK</property>
        <property name="tooltip_text" translatable="yes">Link</property>
        <property name="label" translatable="yes">Link</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child>
      <object class="GtkSeparatorToolItem">
        <property name="draw">False</property>
        <property name="hexpand">True</property>
      </object>
      <packing>
        <property name="expand">True</property>
      </packing>
    </child>
    <child>
      <object class="GtkToolButton">
        <property name="icon-name">edit-clear</property>
        <property name="action-name">ste.CLEAR</property>
        <property name="tooltip_text" translatable="yes">"""
    """Clear Markup</property>
        <property name="label" translatable="yes">Clear Markup</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
  </object>
</interface>
"""
)
FONT_SIZES = [
    8,
    9,
    10,
    11,
    12,
    13,
    14,
    16,
    18,
    20,
    22,
    24,
    26,
    28,
    32,
    36,
    40,
    48,
    56,
    64,
    72,
]

USERCHARS = r"-\w"
PASSCHARS = r"-\w,?;.:/!%$^*&~\"#'"
HOSTCHARS = r"-\w"
PATHCHARS = r"-\w$.+!*(),;:@&=?/~#%"
# SCHEME = "(news:|telnet:|nntp:|file:/|https?:|ftps?:|webcal:)"
SCHEME = "(file:/|https?:|ftps?:|webcal:)"
USER = "[" + USERCHARS + "]+(:[" + PASSCHARS + "]+)?"
HOST = r"([-\w.]+|\[[0-9A-F:]+\])?"
URLPATH = "(/[" + PATHCHARS + "]*)?[^]'.:}> \t\r\n,\\\"]"

(GENURL, HTTP, MAIL, LINK) = list(range(4))


def find_parent_with_attr(self, attr="dbstate"):
    """ """
    # Find a parent with attr:
    obj = self
    while obj:
        if hasattr(obj, attr):
            break
        obj = obj.get_parent()
    return obj


# -------------------------------------------------------------------------
#
# StyledTextEditor
#
# -------------------------------------------------------------------------
class StyledTextEditor(Gtk.TextView):
    """
    StyledTextEditor is an enhanced Gtk.TextView to edit :class:`.StyledText`.

    StyledTextEditor is a gui object for :class:`.StyledTextBuffer`. It offers
    :meth:`set_text` and :meth:`get_text` convenience methods to set and get the
    buffer's text.

    StyledTextEditor provides a formatting toolbar, which can be retrieved
    by the :meth:`get_toolbar` method.

    StyledTextEdit defines a new signal: 'match-changed', which is raised
    whenever the mouse cursor reaches or leaves a matched string in the text.
    The feature uses the regexp pattern mathing mechanism of
    :class:`.StyledTextBuffer`.
    The signal's default handler highlights the URL strings. URL's can be
    followed from the editor's popup menu or by pressing the <CTRL>Left
    mouse button.

    :ivar last_match: previously matched string, used for generating the
                      'match-changed' signal.
    :type last_match: tuple or None
    :ivar match: currently matched string, used for generating the
                 'match-changed' signal.
    :type match: tuple or None
    :ivar spellcheck: spell checker object created for the editor instance.
    :type spellcheck: :class:`.Spell`
    :ivar textbuffer: text buffer assigned to the edit instance.
    :type textbuffer: :class:`.StyledTextBuffer`
    :ivar toolbar: toolbar to be used for text formatting.
    :type toolbar: Gtk.Toolbar
    :ivar url_match: stores the matched URL and other mathing parameters.
    :type url_match: tuple or None

    """

    __gtype_name__ = "StyledTextEditor"

    __gsignals__ = {
        "match-changed": (
            GObject.SignalFlags.RUN_FIRST,
            None,  # return value
            (GObject.TYPE_PYOBJECT,),
        ),  # arguments
    }

    def __init__(self):
        """Setup initial instance variable values."""
        self.textbuffer = UndoableStyledBuffer()
        self.undo_disabled = self.textbuffer.undo_disabled  # see bug 7097
        self.textbuffer.connect("style-changed", self._on_buffer_style_changed)
        self.textbuffer.connect("changed", self._on_buffer_changed)
        self.undo_action = self.redo_action = None
        Gtk.TextView.__init__(self)
        self.set_buffer(self.textbuffer)

        st_cont = self.get_style_context()
        self.linkcolor = get_link_color(st_cont)
        self.textbuffer.linkcolor = self.linkcolor

        self.match = None
        self.last_match = None
        self._init_url_match()
        self.url_match = None

        self.spellcheck = Spell(self)
        self._internal_style_change = False
        self.uimanager = None

        self._connect_signals()

        # variable to not copy to clipboard on double/triple click
        self.selclick = False

    # virtual methods

    def do_match_changed(self, match):
        """
        Default signal handler.

        URL highlighting.

        :param match: the new match parameters
        :type match: tuple or None

        .. warning:: Do not override the handler, but connect to the signal.
        """
        window = self.get_window(Gtk.TextWindowType.TEXT)
        start, end = self.textbuffer.get_bounds()
        self.textbuffer.remove_tag_by_name("hyperlink", start, end)
        if match and (match[MATCH_FLAVOR] in (GENURL, HTTP, MAIL)):
            start_offset = match[MATCH_START]
            end_offset = match[MATCH_END]

            start = self.textbuffer.get_iter_at_offset(start_offset)
            end = self.textbuffer.get_iter_at_offset(end_offset)

            self.textbuffer.apply_tag_by_name("hyperlink", start, end)
            window.set_cursor(HAND_CURSOR)
            self.url_match = match
        elif match and (match[MATCH_FLAVOR] in (LINK,)):
            window.set_cursor(HAND_CURSOR)
            self.url_match = match
        else:
            window.set_cursor(REGULAR_CURSOR)
            self.url_match = None

    def on_insert_at_cursor(self, widget, string):
        """Signal handler. for debugging only."""
        _LOG.debug("Textview insert '%s'" % string)

    def on_delete_from_cursor(self, widget, mode, count):
        """Signal handler. for debugging only."""
        _LOG.debug("Textview delete type %d count %d" % (mode, count))

    def on_paste_clipboard(self, widget):
        """Signal handler. for debugging only."""
        _LOG.debug("Textview paste clipboard")

    def on_motion_notify_event(self, widget, event):
        """
        Signal handler.

        As the mouse cursor moves the handler checks if there's a new
        regexp match at the new location. If match changes the
        'match-changed' signal is raised.
        """
        x, y = self.window_to_buffer_coords(
            Gtk.TextWindowType.WIDGET, int(event.x), int(event.y)
        )
        iter_at_location = self.get_iter_at_location(x, y)
        if isinstance(iter_at_location, tuple):
            iter_at_location = iter_at_location[1]
        self.match = self.textbuffer.match_check(iter_at_location.get_offset())
        tooltip = None
        for tag in (
            tag
            for tag in iter_at_location.get_tags()
            if tag.get_property("name") is not None
            and tag.get_property("name").startswith("link")
        ):
            self.match = (x, y, LINK, tag.data, tag)
            tooltip = self.make_tooltip_from_link(tag)
            break

        if self.match != self.last_match:
            self.emit("match-changed", self.match)

        self.last_match = self.match
        # self.get_root_window().get_pointer()  # Doesn't seem to do anythhing!
        self.set_tooltip_text(tooltip)
        return False

    def make_tooltip_from_link(self, link_tag):
        """
        Return a string useful for a tooltip given a LinkTag object.
        """
        from gramps.gen.simple import SimpleAccess

        win_obj = find_parent_with_attr(self, attr="dbstate")
        display = link_tag.data
        if win_obj:
            simple_access = SimpleAccess(win_obj.dbstate.db)
            url = link_tag.data
            if url.startswith("gramps://"):
                obj_class, prop, value = url[9:].split("/", 2)
                display = simple_access.display(obj_class, prop, value) or url
        return display + (
            (
                _("\nCommand-Click to follow link")
                if mac()
                else _("\nCtrl-Click to follow link")
            )
            if self.get_editable()
            else ""
        )

    def on_button_release_event(self, widget, event):
        """
        Copy selection to clipboard for left click if selection given
        """
        if (
            event.type == Gdk.EventType.BUTTON_RELEASE
            and self.selclick
            and event.button == 1
        ):
            bounds = self.textbuffer.get_selection_bounds()
            if bounds:
                clip = Gtk.Clipboard.get_for_display(
                    Gdk.Display.get_default(), Gdk.SELECTION_PRIMARY
                )
                clip.set_text(
                    str(self.textbuffer.get_text(bounds[0], bounds[1], True)), -1
                )
        return False

    def on_button_press_event(self, widget, event):
        """
        Signal handler.

        Handles the <CTRL> + Left click over a URL match.
        """
        self.selclick = False
        if (
            (event.type == Gdk.EventType.BUTTON_PRESS)
            and (event.button == 1)
            and (self.url_match)
            and (match_primary_mask(event.get_state()) or not self.get_editable())
        ):
            flavor = self.url_match[MATCH_FLAVOR]
            url = self.url_match[MATCH_STRING]
            self._open_url_cb(None, url, flavor)
        elif event.type == Gdk.EventType.BUTTON_PRESS and event.button == 1:
            # on release we will copy selected data to clipboard
            self.selclick = True
        # propagate click
        return False

    def on_populate_popup(self, widget, menu):
        """
        Signal handler.

        Insert extra menuitems:

        1. Insert spellcheck selector submenu for spell checking.
        2. Insert extra menus depending on ULR match result.
        """
        # spell checker submenu
        spell_menu = Gtk.MenuItem(label=_("Spellcheck"))
        spell_menu.set_submenu(self._create_spell_menu())
        spell_menu.show_all()
        menu.prepend(spell_menu)

        search_menu = Gtk.MenuItem(label=_("Search selection on web"))
        search_menu.connect("activate", self.search_web)
        search_menu.show_all()
        menu.append(search_menu)

        # url menu items
        if self.url_match:
            flavor = self.url_match[MATCH_FLAVOR]
            url = self.url_match[MATCH_STRING]

            if flavor == MAIL:
                open_menu = Gtk.MenuItem(label=_("_Send Mail To..."))
                open_menu.set_use_underline(True)
                copy_menu = Gtk.MenuItem(label=_("Copy _E-mail Address"))
                copy_menu.set_use_underline(True)
            else:
                open_menu = Gtk.MenuItem(label=_("_Open Link"))
                open_menu.set_use_underline(True)
                copy_menu = Gtk.MenuItem(label=_("Copy _Link Address"))
                copy_menu.set_use_underline(True)

            if flavor == LINK and self.get_editable():
                edit_menu = Gtk.MenuItem(label=_("_Edit Link"))
                edit_menu.set_use_underline(True)
                edit_menu.connect(
                    "activate",
                    self._edit_url_cb,
                    self.url_match[-1],  # tag
                )
                edit_menu.show()
                menu.prepend(edit_menu)

            copy_menu.connect("activate", self._copy_url_cb, url, flavor)
            copy_menu.show()

            menu.prepend(copy_menu)

            open_menu.connect("activate", self._open_url_cb, url, flavor)
            open_menu.show()
            menu.prepend(open_menu)

    def search_web(self, widget):
        """
        Search the web for selected text.
        """
        selection = self.textbuffer.get_selection_bounds()
        if len(selection) > 0:
            display_url(
                config.get("behavior.web-search-url")
                % {"text": self.textbuffer.get_text(selection[0], selection[1], True)}
            )

    def reset(self):
        """
        Reset the undoable buffer
        """
        self.textbuffer.reset()
        self.undo_action.set_enabled(False)
        self.redo_action.set_enabled(False)

    # private methods

    def _connect_signals(self):
        """Connect to several signals of the super class Gtk.TextView."""
        self.connect("insert-at-cursor", self.on_insert_at_cursor)
        self.connect("delete-from-cursor", self.on_delete_from_cursor)
        self.connect("paste-clipboard", self.on_paste_clipboard)
        self.connect("motion-notify-event", self.on_motion_notify_event)
        self.connect("button-press-event", self.on_button_press_event)
        self.connect("button-release-event", self.on_button_release_event)
        self.connect("populate-popup", self.on_populate_popup)

    def create_toolbar(self, uimanager, window):
        """
        Create a formatting toolbar.

        :returns: toolbar containing text formatting toolitems.
        :rtype: Gtk.Toolbar
        """
        self.uimanager = uimanager
        # build the toolbar
        builder = Gtk.Builder()
        builder.set_translation_domain(glocale.get_localedomain())
        builder.add_from_string(FORMAT_TOOLBAR)

        # define the actions...
        _actions = [
            ("ITALIC", self._on_toggle_action_activate, "<PRIMARY>i", False),
            ("BOLD", self._on_toggle_action_activate, "<PRIMARY>b", False),
            ("UNDERLINE", self._on_toggle_action_activate, "<PRIMARY>u", False),
            ("STRIKETHROUGH", self._on_toggle_action_activate, "<PRIMARY>s", False),
            ("SUPERSCRIPT", self._on_toggle_action_activate, "<PRIMARY>p", False),
            ("SUBSCRIPT", self._on_toggle_action_activate, "<PRIMARY>r", False),
            ("FONTCOLOR", self._on_action_activate),
            ("HIGHLIGHT", self._on_action_activate),
            ("LINK", self._on_link_activate),
            ("CLEAR", self._format_clear_cb),
            ("STUndo", self.undo, "<primary>z"),
            ("STRedo", self.redo, "<primary><shift>z"),
        ]

        # the following are done manually rather than using actions
        fonts = SystemFonts()
        fontface = builder.get_object("Fontface")
        fontface.init(fonts.get_system_fonts(), shortlist=True, validator=None)
        fontface.set_entry_editable(False)
        fontface.connect(
            "changed", make_cb(self._on_valueaction_changed, StyledTextTagType.FONTFACE)
        )
        # set initial value
        default = StyledTextTagType.STYLE_DEFAULT[StyledTextTagType.FONTFACE]
        self.fontface = fontface.get_child()
        self.fontface.set_text(str(default))
        fontface.show()

        items = FONT_SIZES
        fontsize = builder.get_object("Fontsize")
        fontsize.init(items, shortlist=False, validator=is_valid_fontsize)
        fontsize.set_entry_editable(True)
        fontsize.connect(
            "changed", make_cb(self._on_valueaction_changed, StyledTextTagType.FONTSIZE)
        )
        # set initial value
        default = StyledTextTagType.STYLE_DEFAULT[StyledTextTagType.FONTSIZE]
        self.fontsize = fontsize.get_child()
        self.fontsize.set_text(str(default))
        fontsize.show()

        # create the action group and insert all the actions
        self.action_group = ActionGroup("Format", _actions, "ste")
        act_grp = SimpleActionGroup()
        window.insert_action_group("ste", act_grp)
        window.set_application(uimanager.app)
        uimanager.insert_action_group(self.action_group, act_grp)

        self.undo_action = uimanager.get_action(self.action_group, "STUndo")
        self.redo_action = uimanager.get_action(self.action_group, "STRedo")
        # allow undo/redo to see actions if editable.
        self.textbuffer.connect("changed", self._on_buffer_changed)
        # undo/redo are initially greyed out, until something is changed
        self.undo_action.set_enabled(False)
        self.redo_action.set_enabled(False)

        # get the toolbar and set it's style
        toolbar = builder.get_object("ToolBar")

        return toolbar, self.action_group

    def set_transient_parent(self, parent=None):
        self.transient_parent = parent

    def _init_url_match(self):
        """Setup regexp matching for URL match."""
        self.textbuffer.create_tag(
            "hyperlink", underline=Pango.Underline.SINGLE, foreground=self.linkcolor
        )
        self.textbuffer.match_add(
            SCHEME + "//(" + USER + "@)?" + HOST + "(:[0-9]+)?" + URLPATH, GENURL
        )
        self.textbuffer.match_add(
            r"(www\.|ftp\.)["
            + HOSTCHARS
            + r"]*\.["
            + HOSTCHARS
            + ".]+"
            + "(:[0-9]+)?"
            + URLPATH,
            HTTP,
        )
        self.textbuffer.match_add(
            r"(mailto:)?[\w][-.\w]*@[\w]" r"[-\w]*(\.[\w][-.\w]*)+", MAIL
        )

    def _create_spell_menu(self):
        """
        Create a menu with all the installed spellchecks.

        It is called each time the popup menu is opened. Each spellcheck
        forms a radio menu item, and the selected spellcheck is set as active.

        :returns: menu containing all the installed spellchecks.
        :rtype: Gtk.Menu
        """
        active_spellcheck = self.spellcheck.get_active_spellcheck()

        menu = Gtk.Menu()
        group = None
        for lang in self.spellcheck.get_all_spellchecks():
            menuitem = Gtk.RadioMenuItem(label=lang)
            menuitem.set_active(lang == active_spellcheck)
            menuitem.connect("activate", self._spell_change_cb, lang)
            menu.append(menuitem)

            if group is None:
                group = menuitem

        return menu

    # Callback functions

    def _on_toggle_action_activate(self, action, value):
        """
        Toggle a style.

        Toggle styles are e.g. 'bold', 'italic', 'underline', 'strikethrough',
                               'superscript' and 'subscript'.
        """
        action.set_state(value)
        if self._internal_style_change:
            return

        style = action.get_name()
        value = value.get_boolean()

        if value and style in ("SUPERSCRIPT", "SUBSCRIPT"):
            if style == "SUPERSCRIPT":
                toggle_off = "SUBSCRIPT"
            else:
                toggle_off = "SUPERSCRIPT"
            action = self.uimanager.get_action(self.action_group, toggle_off)
            action.change_state(Variant.new_boolean(False))

        _LOG.debug("applying style '%s' with value '%s'" % (style, str(value)))
        self.textbuffer.apply_style(getattr(StyledTextTagType, style), value)

    def _on_link_activate(self, action, value):
        """
        Create a link of a selected region of text.
        """
        # Send in a default link. Could be based on active person.
        selection_bounds = self.textbuffer.get_selection_bounds()
        if selection_bounds:
            # Paste text to clipboards
            text = str(
                self.textbuffer.get_text(selection_bounds[0], selection_bounds[1], True)
            )
            clipboard = Gtk.Clipboard.get_for_display(
                Gdk.Display.get_default(), Gdk.SELECTION_CLIPBOARD
            )

            clipboard.set_text(text, -1)
            clipboard = Gtk.Clipboard.get_for_display(
                Gdk.Display.get_default(), Gdk.SELECTION_PRIMARY
            )
            clipboard.set_text(text, -1)
            uri_dialog(self, None, self.setlink_callback)

    def setlink_callback(self, uri, tag=None):
        """
        Callback for setting or editing a link's object.
        """
        if uri:
            _LOG.debug("applying style 'link' with value '%s'" % uri)
            if not tag:
                tag = LinkTag(
                    self.textbuffer,
                    data=uri,
                    underline=Pango.Underline.SINGLE,
                    foreground=self.linkcolor,
                )
                selection_bounds = self.textbuffer.get_selection_bounds()
                self.textbuffer.apply_tag(tag, selection_bounds[0], selection_bounds[1])
            else:
                tag.data = uri

    def _on_action_activate(self, action, value):
        """Apply a format set from a Gtk.Action type of action."""
        style = getattr(StyledTextTagType, action.get_name())
        current_value = self.textbuffer.get_style_at_cursor(style)

        if style == StyledTextTagType.FONTCOLOR:
            color_dialog = Gtk.ColorChooserDialog(
                title=_("Select font color"), transient_for=self.transient_parent
            )
        elif style == StyledTextTagType.HIGHLIGHT:
            color_dialog = Gtk.ColorChooserDialog(
                title=_("Select background color"), transient_for=self.transient_parent
            )
        else:
            _LOG.debug("unknown style: '%d'" % style)
            return

        if current_value:
            rgba = Gdk.RGBA()
            rgba.parse(current_value)
            color_dialog.set_rgba(rgba)

        response = color_dialog.run()
        rgba = color_dialog.get_rgba()
        value = "#%02x%02x%02x" % (
            int(rgba.red * 255),
            int(rgba.green * 255),
            int(rgba.blue * 255),
        )
        color_dialog.destroy()

        if response == Gtk.ResponseType.OK:
            _LOG.debug("applying style '%d' with value '%s'" % (style, str(value)))
            self.textbuffer.apply_style(style, value)

    def _on_valueaction_changed(self, obj, style):
        """Apply a format set by a ShortListComboEntry."""
        if self._internal_style_change:
            return

        value = obj.get_active_data()
        try:
            value = StyledTextTagType.STYLE_TYPE[style](value)
            _LOG.debug("applying style '%d' with value '%s'" % (style, str(value)))
            self.textbuffer.apply_style(style, value)
        except ValueError:
            _LOG.debug(
                "unable to convert '%s' to '%s'"
                % (value, StyledTextTagType.STYLE_TYPE[style])
            )

    def _format_clear_cb(self, action, value):
        """
        Remove all formats from the selection or from all.

        Remove only our own tags without touching other ones (e.g. Gspell),
        thus remove_all_tags() can not be used.
        """
        clear_anything = self.textbuffer.clear_selection()
        if not clear_anything:
            for style in ALLOWED_STYLES:
                self.textbuffer.remove_style(style)

            start, end = self.textbuffer.get_bounds()
            tags = self.textbuffer._get_tag_from_range(
                start.get_offset(), end.get_offset()
            )
            for tag_name, tag_data in tags.items():
                if tag_name.startswith("link"):
                    for start, end in tag_data:
                        self.textbuffer.remove_tag_by_name(
                            tag_name,
                            self.textbuffer.get_iter_at_offset(start),
                            self.textbuffer.get_iter_at_offset(end + 1),
                        )

    def _on_buffer_changed(self, buffer):
        """synchronize the undo/redo buttons with what is possible"""
        if self.undo_action:
            self.undo_action.set_enabled(self.textbuffer.can_undo)
            self.redo_action.set_enabled(self.textbuffer.can_redo)

    def _on_buffer_style_changed(self, buffer, changed_styles):
        """Synchronize actions as the format changes at the buffer's cursor."""
        if not self.uimanager:
            return  # never initialized a toolbar, not editable
        types = [
            StyledTextTagType.ITALIC,
            StyledTextTagType.BOLD,
            StyledTextTagType.UNDERLINE,
            StyledTextTagType.STRIKETHROUGH,
            StyledTextTagType.SUPERSCRIPT,
            StyledTextTagType.SUBSCRIPT,
        ]
        self._internal_style_change = True
        for style, style_value in changed_styles.items():
            if style in types:
                action = self.uimanager.get_action(
                    self.action_group, StyledTextTagType(style).xml_str().upper()
                )
                action.change_state(Variant.new_boolean(style_value))
            elif style == StyledTextTagType.FONTFACE:
                self.fontface.set_text(style_value)
            elif style == StyledTextTagType.FONTSIZE:
                self.fontsize.set_text(str(style_value))
        self._internal_style_change = False

    def _spell_change_cb(self, menuitem, spellcheck):
        """Set spell checker spellcheck according to user selection."""
        self.spellcheck.set_active_spellcheck(spellcheck)

    def _open_url_cb(self, menuitem, url, flavor):
        """Open the URL in a browser."""
        if not url:
            return

        if flavor == HTTP:
            url = "http:" + url
        elif flavor == MAIL:
            if not url.startswith("mailto:"):
                url = "mailto:" + url
        elif flavor == GENURL:
            pass
        elif flavor == LINK:
            # gramps://person/id/VALUE
            # gramps://person/handle/VALUE
            if url.startswith("gramps://"):
                # if in a window:
                win_obj = find_parent_with_attr(self, attr="dbstate")
                if win_obj:
                    # Edit the object:
                    obj_class, prop, value = url[9:].split("/", 2)
                    from ..editors import EditObject

                    EditObject(
                        win_obj.dbstate,
                        win_obj.uistate,
                        win_obj.track,
                        obj_class,
                        prop,
                        value,
                    )
                    return
        else:
            return
        # If ok, then let's open
        obj = find_parent_with_attr(self, attr="dbstate")
        if obj:
            display_url(url, obj.uistate)
        else:
            display_url(url)

    def _copy_url_cb(self, menuitem, url, flavor):
        """Copy url to both useful selections."""
        clipboard = Gtk.Clipboard.get_for_display(
            Gdk.Display.get_default(), Gdk.SELECTION_CLIPBOARD
        )
        clipboard.set_text(url, -1)

        clipboard = Gtk.Clipboard.get_for_display(
            Gdk.Display.get_default(), Gdk.SELECTION_PRIMARY
        )
        clipboard.set_text(url, -1)

    def _edit_url_cb(self, menuitem, link_tag):
        """
        Edit the URI of the link.
        """
        # Paste text to clipboards
        bounds = self.textbuffer.get_selection_bounds()
        if bounds:
            text = str(self.textbuffer.get_text(bounds[0], bounds[1], True))
            clipboard = Gtk.Clipboard.get_for_display(
                Gdk.Display.get_default(), Gdk.SELECTION_CLIPBOARD
            )
            clipboard.set_text(text, -1)
            clipboard = Gtk.Clipboard.get_for_display(
                Gdk.Display.get_default(), Gdk.SELECTION_PRIMARY
            )
            clipboard.set_text(text, -1)
        uri_dialog(
            self, link_tag.data, lambda uri: self.setlink_callback(uri, link_tag)
        )

    # public methods

    def set_text(self, text):
        """
        Set the text of the text buffer of the editor.

        :param text: formatted text to edit in the view.
        :type text: :class:`.StyledText`
        """
        self.textbuffer.set_text(text)
        self.textbuffer.place_cursor(self.textbuffer.get_start_iter())

    def get_text(self):
        """
        Get the text of the text buffer of the editor.

        :returns: the formatted text from the editor.
        :rtype: :class:`.StyledText`
        """
        start, end = self.textbuffer.get_bounds()
        return self.textbuffer.get_text(start, end, True)

    def undo(self, *obj):
        self.textbuffer.undo()

    def redo(self, *obj):
        self.textbuffer.redo()


# -------------------------------------------------------------------------
#
# Module functions
#
# -------------------------------------------------------------------------


def uri_dialog(self, uri, callback):
    """
    Function to spawn the link editor.
    """
    from ..editors.editlink import EditLink

    obj = find_parent_with_attr(self, attr="dbstate")
    if obj:
        if uri is None:
            # make a default link
            uri = "http://"
            # Check in order for an open page:
            for object_class in [
                "Person",
                "Place",
                "Event",
                "Family",
                "Repository",
                "Source",
                "Media",
            ]:
                handle = obj.uistate.get_active(object_class)
                if handle:
                    uri = "gramps://%s/handle/%s" % (object_class, handle)
                    break
        EditLink(obj.dbstate, obj.uistate, obj.track, uri, callback)


def is_valid_fontsize(size):
    """Validator function for font size selector widget."""
    return (size > 0) and (size < 73)


def make_cb(func, value):
    """
    Generates a callback function based off the passed arguments
    """
    return lambda x: func(x, value)



# encoding:utf-8
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Benny Malengier
# Copyright (C) 2009 Douglas S. Blank
# Copyright (C) 2009 Nick Hall
# Copyright (C) 2011 Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

MODULE_VERSION = "6.0"

# ------------------------------------------------------------------------
#
# default views of Gramps
#
# ------------------------------------------------------------------------

register(
    VIEW,
    id="eventview",
    name=_("Events"),
    description=_("The view showing all the events"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="eventview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    category=("Events", _("Events")),
    viewclass="EventView",
    order=START,
)

register(
    VIEW,
    id="familyview",
    name=_("Families"),
    description=_("The view showing all families"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="familyview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    category=("Families", _("Families")),
    viewclass="FamilyView",
    order=START,
)

register(
    VIEW,
    id="dashboardview",
    name=_("Dashboard"),
    description=_("The view showing Gramplets"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="dashboardview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    category=("Dashboard", _("Dashboard")),
    viewclass="DashboardView",
    order=START,
)

register(
    VIEW,
    id="mediaview",
    name=_("Media"),
    description=_("The view showing all the media objects"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="mediaview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    category=("Media", _("Media")),
    viewclass="MediaView",
    order=START,
)

register(
    VIEW,
    id="noteview",
    name=_("Notes"),
    description=_("The view showing all the notes"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="noteview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    category=("Notes", _("Notes")),
    viewclass="NoteView",
    order=START,
)

register(
    VIEW,
    id="relview",
    name=_("Relationships"),
    description=_("The view showing all relationships of the selected person"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="relview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    category=("Relationships", _("Relationships")),
    viewclass="RelationshipView",
    order=START,
)

register(
    VIEW,
    id="pedigreeview",
    name=_("Pedigree"),
    description=_("The view showing an ancestor pedigree of the selected person"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="pedigreeview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    category=("Ancestry", _("Charts")),
    viewclass="PedigreeView",
    order=START,
    stock_icon="gramps-pedigree",
)

register(
    VIEW,
    id="fanchartview",
    name=_("Fan Chart"),
    category=("Ancestry", _("Charts")),
    description=_("A view showing parents through a fanchart"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="fanchartview.py",
    authors=["Douglas S. Blank", "B. Malengier"],
    authors_email=["doug.blank@gmail.com", "benny.malengier@gmail.com"],
    viewclass="FanChartView",
    stock_icon="gramps-fanchart",
)

register(
    VIEW,
    id="fanchartdescview",
    name=_("Descendant Fan"),
    category=("Ancestry", _("Charts")),
    description=_("Showing descendants through a fanchart"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="fanchartdescview.py",
    authors=["B. Malengier"],
    authors_email=["benny.malengier@gmail.com"],
    viewclass="FanChartDescView",
    stock_icon="gramps-fanchartdesc",
)

register(
    VIEW,
    id="fanchart2wayview",
    name=_("2-Way Fan"),
    category=("Ancestry", _("Charts")),
    description=_("Showing ascendants and descendants through a fanchart"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="fanchart2wayview.py",
    authors=["B. Jacquet"],
    authors_email=["bastien.jacquet_dev@m4x.org"],
    viewclass="FanChart2WayView",
    stock_icon="gramps-fanchart2way",
)

register(
    VIEW,
    id="personview",
    name=_("Grouped People"),
    description=_(
        "The view showing all people in the Family Tree grouped per" " family name"
    ),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="persontreeview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    category=("People", _("People")),
    viewclass="PersonTreeView",
    order=START,
    stock_icon="gramps-tree-group",
)

register(
    VIEW,
    id="personlistview",
    name=_("People"),
    description=_("The view showing all people in the Family Tree" " in a flat list"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="personlistview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    category=("People", _("People")),
    viewclass="PersonListView",
    stock_icon="gramps-tree-list",
)

register(
    VIEW,
    id="placelistview",
    name=_("Places"),
    description=_("The view showing all the places of the Family Tree"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="placelistview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    category=("Places", _("Places")),
    viewclass="PlaceListView",
    stock_icon="gramps-tree-list",
)

register(
    VIEW,
    id="placetreeview",
    name=_("Place Tree"),
    description=_("A view displaying places in a tree format."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="placetreeview.py",
    authors=["Donald N. Allingham", "Gary Burton", "Nick Hall"],
    authors_email=[""],
    category=("Places", _("Places")),
    viewclass="PlaceTreeView",
    stock_icon="gramps-tree-group",
    order=START,
)

register(
    VIEW,
    id="repoview",
    name=_("Repositories"),
    description=_("The view showing all the repositories"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="repoview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    category=("Repositories", _("Repositories")),
    viewclass="RepositoryView",
    order=START,
)

register(
    VIEW,
    id="sourceview",
    name=_("Sources"),
    description=_("The view showing all the sources"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="sourceview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    category=("Sources", _("Sources")),
    viewclass="SourceView",
    order=START,
    stock_icon="gramps-tree-list",
)

register(
    VIEW,
    id="citationlistview",
    name=_("Citations"),
    description=_("The view showing all the citations"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="citationlistview.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
    category=("Citations", _("Citations")),
    viewclass="CitationListView",
    order=START,
)

register(
    VIEW,
    id="citationtreeview",
    name=_("Citation Tree"),
    description=_("A view displaying citations and sources in a tree format."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="citationtreeview.py",
    authors=["Tim G L Lyons", "Nick Hall"],
    authors_email=[""],
    category=("Sources", _("Sources")),
    viewclass="CitationTreeView",
    stock_icon="gramps-tree-select",
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Set
from ....lib import Person
from ....db import Database
from ....types import PersonHandle


# -------------------------------------------------------------------------
#
# IsMoreThanNthGenerationAncestorOf
#
# -------------------------------------------------------------------------
class IsMoreThanNthGenerationAncestorOf(Rule):
    """Rule that checks for a person that is an ancestor of a specified person
    at least N generations away"""

    labels = [_("ID:"), _("Number of generations:")]
    name = _("Ancestors of <person> at least <N> generations away")
    category = _("Ancestral filters")
    description = _(
        "Matches people that are ancestors "
        "of a specified person at least N generations away"
    )

    def prepare(self, db: Database, user):
        self.db = db
        self.selected_handles: Set[PersonHandle] = set()
        person = db.get_person_from_gramps_id(self.list[0])
        if person:
            root_handle = person.handle
            if root_handle:
                self.init_ancestor_list(root_handle)

    def init_ancestor_list(self, root_handle: PersonHandle):
        queue = [(root_handle, 1)]  # generation 1 is root
        while queue:
            handle, gen = queue.pop(0)  # pop off front of queue
            if gen > int(self.list[1]):
                self.selected_handles.add(handle)
            gen += 1
            p = self.db.get_person_from_handle(handle)
            fam_id = p.parent_family_list[0] if len(p.parent_family_list) > 0 else None
            if fam_id:
                fam = self.db.get_family_from_handle(fam_id)
                if fam:
                    f_id = fam.father_handle
                    m_id = fam.mother_handle
                    # append to back of queue:
                    if f_id:
                        queue.append((f_id, gen))
                    if m_id:
                        queue.append((m_id, gen))

    def reset(self):
        self.selected_handles.clear()

    def apply_to_one(self, db, person: Person) -> bool:
        return person.handle in self.selected_handles



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import logging

_LOG = logging.getLogger(".widgets.statusbar")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from . import WarnButton


# -------------------------------------------------------------------------
#
# Statusbar class
#
# -------------------------------------------------------------------------
class Statusbar(Gtk.Box):
    """
    A status bar
    """

    def __init__(self):
        Gtk.Box.__init__(self)
        self.set_spacing(4)
        self.set_border_width(2)

        self.__progress = Gtk.ProgressBar()
        self.__progress.set_valign(Gtk.Align.CENTER)
        self.__progress.set_size_request(100, -1)
        self.__progress.hide()

        self.__warnbtn = WarnButton()

        self.__version = Gtk.Button()
        self.__version.set_relief(Gtk.ReliefStyle.NONE)
        self.__version.get_style_context().add_class("destructive-action")

        self.__status = Gtk.Statusbar()
        self.__status.show()

        self.__filter = Gtk.Label()
        self.__filter.set_halign(Gtk.Align.END)
        self.__filter.show()

        self.pack_start(self.__warnbtn, False, True, 4)
        self.pack_start(self.__progress, False, True, 4)
        self.pack_start(self.__status, True, True, 4)
        self.pack_end(self.__filter, False, True, 4)
        self.pack_end(self.__version, False, True, 4)

    def get_warning_button(self):
        """Return the warning button widget."""
        return self.__warnbtn

    def get_progress_bar(self):
        """Return the progress bar widget."""
        return self.__progress

    def get_version_btn(self):
        """Return the version button widget."""
        return self.__version

    def get_context_id(self, context_description):
        """Return a new or existing context identifier."""
        return self.__status.get_context_id(context_description)

    def push(self, context_id, text):
        """Push message onto a statusbar's stack."""
        return self.__status.push(context_id, text)

    def pop(self, context_id):
        """Remove the top message from a statusbar's stack."""
        self.__status.pop(context_id)

    def remove(self, context_id, message_id):
        """Remove the message with the specified message_id."""
        self.__status.remove(context_id, message_id)

    def set_filter(self, text):
        """Set the filter status text."""
        self.__filter.set_text(text)

    def clear_filter(self):
        """Clear the filter status text."""
        self.__filter.set_text("")
